# coding: utf-8
from __future__ import absolute_import, unicode_literals

import types
import six

from bisect import insort
from thebot import Plugin


class Plugin(Plugin):
    """Allows other plugins to notify user via different adapters.

    This plugin has no it's own commands but you can control
    it's behaviour via notification-priorities settings, like that:

    set notification-priorities xmpp,irc,email
    """
    deps = ['settings']

    def notify(self, identity, message):
        identity_plugin = self.bot.get_plugin('identity')
        settings_plugin = self.bot.get_plugin('settings')

        if isinstance(identity, six.string_types):
            identity = identity_plugin.get_identity_by_id(identity)

        priorities = settings_plugin.get(identity.id, 'notification-priorities', '')
        priorities = enumerate(item.strip() for item in priorities.split(','))
        priorities = dict((key, value) for value, key in priorities)
        # now priorities is a map from adapter's name to a number

        # there we'll store sorted by priority
        online_contacts = []

        for default_priority, contact in enumerate(identity.persons, 1000):
            # we need increasing default_priority, to keep order of
            # contacts without preferences

            if contact.adapter.is_online(contact.user):
                insort(
                    online_contacts,
                    (
                        priorities.get(contact.adapter.name, default_priority),
                        contact
                    )
                )

        if online_contacts:
            contact = online_contacts[0][1]
            contact.adapter.send(message, contact.user)
            return True

        return False




# coding: utf-8
from __future__ import absolute_import, unicode_literals

import argparse
import importlib
import logging
import os
import pickle
import pkg_resources
import re
import server_reloader
import shelve
import six
import textwrap
import threading
import time
import yaml

from .utils import MutableMapping, force_str, printable

__version__ = pkg_resources.get_distribution(__name__).version


# pass this object to callback, to terminate the bot
EXIT = object()


@printable
class User(object):
    def __init__(self, id):
        self.id = id

    def __unicode__(self):
        return self.id

    def __eq__(self, another):
        return self.id == another.id


@printable
class Room(object):
    def __init__(self, id):
        self.id = id

    def __unicode__(self):
        return self.id

    def __eq__(self, another):
        return self.id == another.id


@printable
class Request(object):
    def __init__(self, adapter, message, user, room=None, refer_by_name=False):
        self.adapter = adapter
        self.message = message
        self.user = user
        self.room = room
        self.refer_by_name = refer_by_name

    def __unicode__(self):
        result = '{} from {}'.format(self.message, self.user)
        if self.room:
            result += ' at {}'.format(self.room)
        return result

    def respond(self, message):
        self.adapter.send(
            message,
            user=self.user,
            room=self.room,
            refer_by_name=self.refer_by_name,
        )

    def shout(self, message):
        """This method should be used to say something to the channel or a chatroom."""
        self.adapter.send(
            message,
            room=self.room,
        )



@printable
class Adapter(object):
    def __init__(self, bot, callback):
        self.bot = bot
        self.callback = callback

    def __unicode__(self):
        return self.name

    def start(self):
        """You have to override this method, if you plugin requires some background activity.

        Here you can create a thread, but don't forget to make its `daemon` attrubute equal to True.
        """

    def is_online(self, user):
        return False


@printable
class Plugin(object):
    def __init__(self, bot):
        self.bot = bot
        self.storage = self.bot.storage.with_prefix(self.__module__ + ':')
        self.logger = logging.getLogger('thebot.plugin.' + self.name)

    def __unicode__(self):
        return self.name

    def get_callbacks(self):
        for name in dir(self):
            value = getattr(self, name)
            if callable(value):
                patterns = getattr(value, '_patterns', [])
                for pattern in patterns:
                    pattern.plugin_name = self.name
                    yield (pattern, value)


class ThreadedPlugin(Plugin):
    """ThreadedPlugin allows you to do some processing in a background thread.

    This class will take care on proper thread execution and termination.

    * First, implement method `do_job`, which will be executed with given interval.
    * Then run method `self.start_worker(interval=60)` to run a thread.
      It will call your do_job callback each 60 seconds.
    * To stop job execution, call `self.stop_worker()`.

    See `thebot-instagram`, as an example.
    """
    def do_job(self):
        raise NotImplemented('Implement "do_job" method to get real work done.')

    def is_working(self):
        thread = getattr(self, '_thread', None)
        return thread is not None and thread.is_alive()

    def start_worker(self, interval=60):
        if self.is_working():
            return

        self._event = threading.Event()
        self._thread = threading.Thread(target=self._worker, kwargs=dict(interval=interval))
        self._thread.daemon = True
        self._thread.start()

    def stop_worker(self, wait=True):
        event = getattr(self, '_event', None)
        if event is not None:
            event.set()

        if wait:
            self._thread.join()

    def _worker(self, interval=60):
        countdown = 0
        logger = logging.getLogger('thebot.' + self.__class__.__name__)

        on_start = getattr(self, 'on_start', None)
        if on_start is not None:
            on_start()

        while not self._event.is_set():
            if countdown == 0:
                try:
                    self.do_job()
                except Exception:
                    logger.exception('Error during the task execution')

                countdown = interval
            else:
                countdown -= 1

            time.sleep(1)

        on_stop = getattr(self, 'on_stop', None)
        if on_stop is not None:
            on_stop()


@printable
class Re(object):
    def __init__(self, pattern):
        self.pattern = pattern
        self._re = None

    def __unicode__(self):
        return self.pattern

    def match(self, message):
        if self._re is not None:
            return self._re.match(message)


class PatternRe(Re):
    def __init__(self, pattern):
        super(PatternRe, self).__init__(pattern)
        self._re = re.compile('.*' + self.pattern + '.*')


class CommandRe(Re):
    def __init__(self, pattern):
        super(CommandRe, self).__init__(pattern)
        self._re = re.compile('^' + pattern + '$')


def _make_routing_decorator(pattern_cls):
    def decorator(pattern):
        """Decorator to assign routes to plugin's methods.
        """
        def deco(func):
            if getattr(func, '_patterns', None) is None:
                func._patterns = []
            func._patterns.append(pattern_cls(pattern))
            return func
        return deco
    return decorator


on_pattern = _make_routing_decorator(PatternRe)
on_command = _make_routing_decorator(CommandRe)


class HelpPlugin(Plugin):
    """Shows help and basic information about TheBot."""
    name = 'help'

    def __init__(self, *args, **kwargs):
        super(HelpPlugin, self).__init__(*args, **kwargs)
        self._started_at = time.time()

    @on_command('help')
    def show_commands(self, request):
        """Show available commands."""

        lines = ['This is the list of available plugins:']

        def gen_line(plugin):
            if plugin.__doc__:
                return '  * {} — {}'.format(plugin.name, plugin.__doc__.split('\n')[0])
            else:
                return '  * {}'.format(plugin.name)

        lines.extend(sorted(map(gen_line, self.bot.plugins)))

        lines.append('Use \'help <plugin>\' to learn about each plugin.')
        request.respond('\n'.join(lines))


    @on_command('help (?P<plugin>.*)')
    def help(self, request, plugin):
        """Show help on given plugin."""
        lines = []
        commands = []
        reactions = []

        def gen_docs(pattern_list):
            previous_callback = None

            for pattern, callback in pattern_list:
                docstring = utils.force_unicode(callback.__doc__)
                if not docstring:
                    docstring = 'No description'

                if previous_callback != callback:
                    lines.append('    {} — {}'.format(pattern.pattern, docstring))
                else:
                    lines.append('    ' + pattern.pattern)

                previous_callback = callback

        try:
            plugin = self.bot.get_plugin(plugin)
        except KeyError as e:
            request.respond(''.join(e.args))
            return

        # dividing all patterns to commands and hearings
        for pattern, callback in plugin.get_callbacks():
            if isinstance(pattern, CommandRe):
                commands.append((pattern, callback))
            else:
                reactions.append((pattern, callback))

        if plugin.__doc__:
            splitted = plugin.__doc__.split('\n', 1)
            doc = ' — {}'.format(splitted[0])

            if len(splitted) == 2:
                remaining = splitted[1]
                remaining = remaining.rstrip()
                remaining = textwrap.dedent(remaining)
                doc += '\n{}'.format(remaining)
        else:
            doc = '.'

        lines.append('Plugin \'{}\'{}\n'.format(plugin.name, doc))

        if commands:
            lines.append('Provides following commands:')
            gen_docs(commands)

        if reactions:
            if lines:
                lines.append('')

            lines.append('And reacts on these patterns:')
            gen_docs(reactions)

        request.respond('\n'.join(lines))

    @on_command('version')
    def version(self, request):
        """Shows TheBot's version."""
        request.respond('My version is "{}".'.format(__version__))

    @on_command('uptime')
    def uptime(self, request):
        """Shows TheBot's uptime."""
        uptime = time.time() - self._started_at
        days = int(uptime / (24 * 3600))
        hours = int((uptime % (24 * 3600) / 3600))
        minutes = int((uptime % 3600 / 60))
        seconds = int((uptime % 60))

        uptime ='My uptime is '
        if days:
            uptime += '{} day(s)'.format(days)
        elif hours:
            uptime += '{} hour(s)'.format(hours)
        elif minutes:
            uptime += '{} minute(s).'.format(minutes)
        else:
            uptime += '{} second(s).'.format(seconds)

        request.respond(uptime)


@printable
class Stub(object):
    """A stub class to replace objects which can't be unpickled.

    It allows to call any method or access any attribute.
    """
    def __init__(self, name):
        self.name = name

    def __getattr__(self, name):
        return Stub(self.name + '.' + name)

    def __call__(self, *args, **kwargs):
        return None

    def __unicode__(self):
        return self.name


class Pickler(pickle.Pickler):
    def __init__(self, file, protocol=None, global_objects=None):
        pickle.Pickler.__init__(self, file, protocol=protocol)
        self._global_objects = dict(
            (id(obj), persistent_id)
            for persistent_id, obj in (global_objects or {}).items()
        )

    def persistent_id(self, obj):
        if isinstance(obj, Stub):
            return obj.name
        return self._global_objects.get(id(obj))


class Unpickler(pickle.Unpickler):
    """A custom unpickler, to restore references to adapters, plugins and the bot."""
    def __init__(self, file, global_objects=None):
        pickle.Unpickler.__init__(self, file)
        self._global_objects = global_objects or {}

    def persistent_load(self, obj_id):
        try:
            return self._global_objects[obj_id]
        except KeyError:
            return Stub(obj_id)


class Shelve(shelve.DbfilenameShelf):
    """A custom Shelve, to use custom Pickler and Unpickler.
    """
    def __init__(self, filename, global_objects=None):
        shelve.DbfilenameShelf.__init__(self, filename)
        self._global_objects = global_objects or {}

    def __getitem__(self, key):
        try:
            value = self.cache[key]
        except KeyError:
            f = six.BytesIO(self.dict[key])
            value = Unpickler(f, global_objects=self._global_objects).load()
            if self.writeback:
                self.cache[key] = value
        return value

    def __setitem__(self, key, value):
        if self.writeback:
            self.cache[key] = value
        f = six.BytesIO()
        p = Pickler(f, self._protocol, global_objects=self._global_objects)
        p.dump(value)
        value = f.getvalue()
        self.dict[key] = value


class Storage(utils.MutableMapping):
    def __init__(self, filename, prefix='', global_objects=None):
        """Specials are used to restore references to some nonserializable objects,
        such as TheBot itself.
        """
        if isinstance(filename, Shelve):
            self._shelve = filename
        else:
            self._shelve = Shelve(filename, global_objects=global_objects)

        self.prefix = prefix
        self.global_objects = global_objects or {}

    def __getitem__(self, name):
        return self._shelve.__getitem__(utils.force_str(self.prefix + name))

    def __setitem__(self, name, value):
        return self._shelve.__setitem__(utils.force_str(self.prefix + name), value)

    def __delitem__(self, name):
        return self._shelve.__delitem__(utils.force_str(self.prefix + name))

    def __len__(self):
        return sum(1 for key in self)

    def __iter__(self):
        prefix_len = len(self.prefix)
        return (
            key[prefix_len:]
            for key in self._shelve.keys()
                if key.startswith(self.prefix)
        )

    def keys(self):
        return list(self)

    def clear(self):
        for key in self.keys():
            del self[key]

    def with_prefix(self, prefix):
        return Storage(self._shelve, prefix=self.prefix + prefix, global_objects=self.global_objects)

    def close(self):
        self._shelve.close()


class Config(object):
    def __init__(self):
        self._data = {}

    def read_from_file(self, filename):
        with open(filename) as f:
            self.read_from_string(f.read())

    def read_from_string(self, data):
        data = yaml.load(data)

        # This complex code allows you
        # to translate such YAML config:
        #
        # adapters: [xmpp, http]
        # xmpp:
        #   jid: thebot@ya.ru
        # http:
        #   port: 9000
        #
        # into a config with attributes `xmpp_jid` and `http_port`.
        for key in ('plugins', 'adapters'):
            for item in data.get(key, []):
                item_settings = data.pop(item, None)
                if item_settings is not None:
                    for k, v in item_settings.items():
                        data[item + '_' + k] = v


        self._data.update(data)

    def update_from_dict(self, args):
        """Gets settings from the Namespace object, generated by argparse."""
        for key, value in args.items():
            if value is not None:
                if key in ('adapters', 'plugins'):
                    value = value.split(',')

                self._data[key] = value

    def __repr__(self):
        return 'Config: {}'.format(self._data)

    def __getattr__(self, name):
        head = name
        tail = ''
        data = self._data
        
        while head:
            if head in data:
                data = data[head]
                head = tail
                tail = ''
            else:
                if '_' in head:
                    head, part = head.rsplit('_', 1)
                    tail = part + '_' + tail
                else:
                    raise AttributeError('Attribute {0} does not exist'.format(name))
        return data


class Bot(object):
    def __init__(
            self,
            command_line_args=[],
            adapters=None,
            plugins=None,
            config_dict={},
            config_filename='thebot.conf',
        ):

        self.adapters = []
        self.plugins = []
        self.patterns = []
        self.exiting = False

        def create_loader(cls='Adapter'):
            def load(name):
                """Returns class by it's name.

                Given a 'irc' string it will try to load the following:

                1) from thebot_irc import Adapter
                2) from thebot.batteries.irc import Adapter

                If all of them fail, it will raise ImportError
                """

                if isinstance(name, six.string_types):
                    try:
                        module = importlib.import_module('thebot_' + name)
                    except ImportError:
                        module = importlib.import_module('thebot.batteries.' + name)

                    value = getattr(module, cls)
                    if not hasattr(value, 'name'):
                        value.name = name
                else:
                    value = name

                return value

            loaded = set()

            def loader(names):
                """Recursive loader."""
                for name in names:
                    if name not in loaded:
                        cls = load(name)
                        loaded.add(name)

                        deps = getattr(cls, 'deps', ())
                        for dep in loader(deps):
                            yield dep
                        yield cls

            return loader


        default_config = """
            unittest: no
        """

        # read config first time, to figure out possible adapter
        config = Config()
        config.read_from_string(default_config)

        parser = Bot.get_general_options()
        defaults = dict(parser.parse_args([])._get_kwargs())
        config.update_from_dict(defaults)

        if os.path.exists(config_filename):
            config.read_from_file(config_filename)

        args, unknown = parser.parse_known_args(
            list(filter(lambda x: x not in ('--help', '-h'), command_line_args))
        )
        # we override options only if there was specified value different from default
        args = dict(
            (key, value)
            for key, value in args._get_kwargs()
                if value != defaults.get(key)
        )
        config.update_from_dict(args)


        # now, load plugin and adapter classes, collect their options
        # and parse command line again
        load_plugins = create_loader('Plugin')
        load_adapters = create_loader('Adapter')

        if adapters is None:
            adapters = config.adapters
        adapter_classes = list(load_adapters(adapters))

        if plugins is None:
            plugins = config.plugins

        plugin_classes = list(load_plugins(plugins))
        plugin_classes.append(HelpPlugin)

        for cls in adapter_classes + plugin_classes:
            if hasattr(cls, 'get_options'):
                cls.get_options(parser)


        # now, reread config to write there all defaults from all plugins
        config = Config()
        config.read_from_string(default_config)

        # getting options' defaults from parser
        defaults = dict(parser.parse_args([])._get_kwargs())
        config.update_from_dict(defaults)

        # now applying settings from the config file
        if os.path.exists(config_filename):
            config.read_from_file(config_filename)

        # and finally, override them with command line options
        # if there was specified value different from default
        args = parser.parse_args(command_line_args)
        args = dict(
            (key, value)
            for key, value in args._get_kwargs()
                if value != defaults.get(key)
        )
        config.update_from_dict(args)
        self.config = config

        for key, value in config_dict.items():
            setattr(self.config, key, value)

        logging.basicConfig(
            filename=self.config.log_filename,
            format='[%(asctime)s] %(levelname)s %(name)s: %(message)s',
            level=logging.DEBUG if self.config.verbose else logging.WARNING,
        )

        with open(self.config.pid_filename, 'w') as f:
            f.write(str(os.getpid()))

        # adapters and plugins initialization
        global_objects = dict(bot=self)

        for adapter in adapter_classes:
            a = adapter(self, callback=self.on_request)
            global_objects[a.name] = a
            a.start()
            self.adapters.append(a)

        self.storage = Storage(self.config.storage_filename, global_objects=global_objects)

        for plugin_cls in plugin_classes:
            p = plugin_cls(self)
            self.plugins.append(p)
            callbacks = p.get_callbacks()
            self.patterns.extend(callbacks)

        if self.config.reload_on_changes:
            server_reloader.trigger_on_code_changes()

    @staticmethod
    def get_general_options():
        parser = argparse.ArgumentParser(
            description='The Bot — Hubot\'s killer.'
        )
        parser.add_argument(
            '--verbose', '-v', action='store_true', default=False,
            help='Show more output. Default: False.'
        )
        parser.add_argument(
            '--log-filename', default='thebot.log',
            help='Log\'s filename. Default: thebot.log.'
        )
        parser.add_argument(
            '--pid-filename', default='thebot.pid',
            help='TheBot\'s pid filename. Default: thebot.pid.'
        )
        parser.add_argument(
            '--storage-filename', default='thebot.storage',
            help='Path to a database file, used for TheBot\'s memory. Default: thebot.storage.'
        )
        parser.add_argument(
            '--reload-on-changes', action='store_true', default=False,
            help='Track source files changes and restart the bot. Default: False.'
        )

        group = parser.add_argument_group('General options')
        group.add_argument(
            '--adapters', '-a', default='console',
            help='Adapters to use. You can specify a comma-separated list to use more than one adapter. Default: console.',
        )
        group.add_argument(
            '--plugins', '-p', default='image,math,todo',
            help='Plugins to use. You can specify a comma-separated list to use more than one plugin. Default: image.',
        )
        return parser

    def on_request(self, request, direct=True):
        if request is EXIT:
            self.exiting = True
        else:
            for pattern, callback in self.patterns:
                match = pattern.match(request.message)
                if match is not None:
                    try:
                        result = callback(request, **match.groupdict())
                    except Exception:
                        logging.getLogger('thebot.core.on_request').exception(
                            'During processing "{0}" request'.format(request))
                    else:
                        if result is not None:
                            raise RuntimeError('Plugin {0} should not return response directly. Use request.respond(some message).')
                    break
            else:
                if direct:
                    # If message wass addressed to TheBot, then it
                    # should report that he does not know such command.
                    request.respond('I don\'t know command "{0}".'.format(request.message))

    def close(self):
        """Will close all connections here.
        """
        self.storage.close()

    def get_adapter(self, name):
        """Returns adapter by it's name."""
        for adapter in self.adapters:
            if getattr(adapter, 'name', None) == name:
                return adapter
        raise KeyError('Adapter {} not found.'.format(name))

    def get_plugin(self, name):
        """Returns plugin by it's name."""
        for plugin in self.plugins:
            if getattr(plugin, 'name', None) == name:
                return plugin
        raise KeyError('Plugin {} not found.'.format(name))




# coding: utf-8

from __future__ import absolute_import, unicode_literals

import times
import datetime
import mock
import thebot
import sys
import re

from thebot import Request, User, Adapter, Plugin, Storage, Config, on_pattern, on_command, Stub
from thebot.batteries import todo
from thebot.batteries.identity import Person
from nose.tools import eq_, assert_raises
from contextlib import closing

PYTHON_VERSION = '-'.join(map(str, sys.version_info[:3]))
STORAGE_FILENAME = 'unittest-{}.storage'.format(PYTHON_VERSION)


class Bot(thebot.Bot):
    """Test bot which uses slightly different settings."""
    def __init__(self, *args, **kwargs):
        kwargs['config_dict'] = dict(
            unittest=True,
            log_filename='unittest.log',
            storage_filename=STORAGE_FILENAME,
        )
        kwargs['config_filename'] = 'unexistent.conf'
        super(Bot, self).__init__(*args, **kwargs)

    def close(self):
        self.storage.clear()
        super(Bot, self).close()


class TestAdapter(Adapter):
    name = 'test'

    def __init__(self, *args, **kwargs):
        super(TestAdapter, self).__init__(*args, **kwargs)
        self._lines = []
        self._offline_users = set()

    def send(self, message, user=None, room=None, refer_by_name=None):
        self._lines.append(message)

    def write(self, input_line, user='some user'):
        """This method is for test purpose.
        """
        name = 'Thebot, '

        if input_line.startswith(name):
            self.callback(Request(self, input_line[len(name):], user=User(user)), direct=True)
        else:
            self.callback(Request(self, input_line, user=User(user)), direct=False)

    def offline(self, user):
        self._offline_users.add(user)

    def is_online(self, user):
        return not user.id in self._offline_users


class TestPlugin(Plugin):
    """A simple test plugin.

    With extended documentation.
    Which can be multiline, and will be shown as plugin's help.
    """
    name = 'test'
    @on_pattern('cat')
    def i_like_cats(self, request):
        """Shows how TheBot likes cats."""
        request.respond('I like cats!!!')

    @on_command('search (?P<this>.*)')
    @on_command('find (?P<this>.*)')
    def find(self, request, this=None):
        """Making a fake search of the term."""
        request.respond('I found {0}'.format(this))


def test_install_adapters():
    with closing(Bot(adapters=[TestAdapter], plugins=[])) as bot:
        assert len(bot.adapters) == 1


def test_install_plugins():
    with closing(Bot(adapters=[], plugins=[TestPlugin])) as bot:
        eq_(0, len(bot.adapters))
        eq_(2, len(bot.plugins)) # Help plugin is added by default
        eq_(7, len(bot.patterns))


def test_one_line():
    with closing(Bot(adapters=[TestAdapter], plugins=[TestPlugin])) as bot:
        adapter = bot.get_adapter('test')

        eq_(adapter._lines, [])
        adapter.write('I have a cat')
        eq_(adapter._lines, ['I like cats!!!'])

        adapter.write('find Umputun')
        eq_(adapter._lines[-1], 'I found Umputun')


def test_unknown_command():
    """TheBot reports about unknown commands, addressed directly to him."""

    with closing(Bot(adapters=[TestAdapter], plugins=[TestPlugin])) as bot:
        adapter = bot.adapters[0]

        eq_(adapter._lines, [])
        adapter.write('Thebot, some command')
        eq_(adapter._lines, ['I don\'t know command "some command".'])


def test_exception_raised_if_plugin_returns_not_none():
    class BadPlugin(Plugin):
        name = 'bad'
        @on_command('do')
        def do(self, request):
            return 'Hello world'


    with closing(Bot(adapters=[TestAdapter], plugins=[BadPlugin])) as bot:
        adapter = bot.adapters[0]

        assert_raises(RuntimeError, adapter.write, 'do')


def test_simple_storage():
    with closing(Storage(STORAGE_FILENAME)) as storage:
        storage.clear()

        eq_([], storage.keys())

        storage['blah'] = 'minor'
        storage['one'] = {'some': 'dict'}

        eq_(['blah', 'one'], sorted(storage.keys()))
        eq_('minor', storage['blah'])


def test_storage_nesting():
    with closing(Storage(STORAGE_FILENAME)) as storage:
        storage.clear()

        first = storage.with_prefix('first:')
        second = storage.with_prefix('second:')

        eq_([], storage.keys())

        first['blah'] = 'minor'
        second['one'] = {'some': 'dict'}

        eq_(['first:blah', 'second:one'], sorted(storage.keys()))
        eq_(['first:blah', 'second:one'], sorted(list(storage)))

        eq_(['blah'], first.keys())
        eq_(['blah'], list(first))

        eq_(['one'], second.keys())
        eq_(['one'], list(second))

        eq_('minor', first['blah'])
        assert_raises(KeyError, lambda: second['blah'])

        first.clear()
        eq_(['second:one'], sorted(storage.keys()))
        eq_(['second:one'], sorted(list(storage)))

def test_storage_deep_nesting():
    with closing(Storage(STORAGE_FILENAME)) as storage:
        storage.clear()

        first = storage.with_prefix('first:')
        second = first.with_prefix('second:')

        second['blah'] = 'minor'

        eq_(['first:second:blah'], storage.keys())
        eq_(['second:blah'], first.keys())
        eq_(['blah'], second.keys())


def test_help_command():
    with closing(Bot(adapters=[TestAdapter], plugins=[TestPlugin])) as bot:
        adapter = bot.adapters[0]

        adapter.write('help')
        eq_(
            [
                'This is the list of available plugins:\n'
                '  * help — Shows help and basic information about TheBot.\n'
                '  * test — A simple test plugin.\n'
                'Use \'help <plugin>\' to learn about each plugin.'
            ],
            adapter._lines
        )


def test_help_plugin_command():
    with closing(Bot(adapters=[TestAdapter], plugins=[TestPlugin])) as bot:
        adapter = bot.adapters[0]

        adapter.write('help test')
        eq_(
            [
                'Plugin \'test\' — A simple test plugin.\n'
                '\n'
                'With extended documentation.\n'
                'Which can be multiline, and will be shown as plugin\'s help.\n'
                '\n'
                'Provides following commands:\n'
                '    find (?P<this>.*) — Making a fake search of the term.\n'
                '    search (?P<this>.*)\n'
                '\n'
                'And reacts on these patterns:\n'
                '    cat — Shows how TheBot likes cats.'
            ],
            adapter._lines
        )


def test_help_plugin_not_found():
    with closing(Bot(adapters=[TestAdapter], plugins=[TestPlugin])) as bot:
        adapter = bot.adapters[0]

        adapter.write('help blah')
        eq_('Plugin blah not found.', adapter._lines[-1])


def test_delete_from_storage():
    with closing(Storage(STORAGE_FILENAME)) as storage:
        storage.clear()

        storage['blah'] = 'minor'
        del storage['blah']

        eq_([], sorted(storage.keys()))


def test_storage_is_iterable_as_dict():
    with closing(Storage(STORAGE_FILENAME)) as storage:
        storage.clear()

        storage['blah'] = 'minor'
        storage['another'] = 'option'

        eq_(['another', 'blah'], sorted(storage.keys()))
        eq_(['minor', 'option'], sorted(storage.values()))
        eq_([('another', 'option'), ('blah', 'minor')], sorted(storage.items()))

        eq_(
            ['another', 'blah'],
            sorted(item for item in storage)
        )



def test_storage_restores_bot_attribute():
    with closing(Bot(adapters=[TestAdapter], plugins=[TestPlugin])) as bot:
        adapter = bot.get_adapter('test')
        original = Request(adapter, 'some text', 'a user')

        bot.storage['request'] = original

        restored = bot.storage['request']
        eq_(restored.adapter, original.adapter)


def test_storage_with_prefix_keeps_global_objects():
    with closing(Storage(STORAGE_FILENAME, global_objects=dict(some='value'))) as storage:
        prefixed = storage.with_prefix('nested:')

        eq_(storage.global_objects, prefixed.global_objects)


def test_get_adapter_by_name():
    with closing(Bot(adapters=[TestAdapter])) as bot:
        adapter = bot.get_adapter('test')
        assert isinstance(adapter, TestAdapter)


def test_todo_plugin():
    with closing(Bot(adapters=[TestAdapter], plugins=[todo.Plugin])) as bot:
        adapter = bot.get_adapter('test')

        adapter.write('remind at 2012-10-05 to Celebrate my birthday')
        adapter.write('remind at 2012-12-18 to Celebrate daughter\'s birthday')
        adapter.write('remind at 2012-09-01 to Write a doc for TheBot')
        adapter.write('my tasks')

        eq_(
            [
                'ok',
                'ok',
                'ok',
                '16) 2012-09-01 00:00 Write a doc for TheBot\n'
                'cd) 2012-10-05 00:00 Celebrate my birthday\n'
                '9c) 2012-12-18 00:00 Celebrate daughter\'s birthday',
            ],
            adapter._lines
        )


def test_todo_plugin_for_different_users():
    with closing(Bot(adapters=[TestAdapter], plugins=[todo.Plugin])) as bot:
        adapter = bot.get_adapter('test')

        adapter.write('remind at 2012-10-05 to Celebrate my birthday', user='blah')
        adapter.write('remind at 2012-12-18 to Celebrate daughter\'s birthday', user='minor')

        adapter._lines[:] = []
        adapter.write('my tasks', user='minor')

        eq_(
            [
                '9c) 2012-12-18 00:00 Celebrate daughter\'s birthday',
            ],
            adapter._lines
        )


def test_todo_remind():
    with closing(Bot(adapters=[TestAdapter], plugins=[todo.Plugin])) as bot:
        adapter = bot.get_adapter('test')
        plugin = bot.get_plugin('todo')


        adapter.write('set timezone Asia/Shanghai')
        # these are the local times
        adapter.write('remind at 2012-09-05 10:00 to do task1')
        adapter.write('remind at 2012-09-05 10:30 to do task2')

        with mock.patch.object(times, 'now') as now:
            # this is a server time in UTC
            # it is 10:01 at Shanghai (+8 hours)
            now.return_value = datetime.datetime(2012, 9, 5, 2, 1)

            adapter._lines[:] = []
            plugin._remind_users_about_their_tasks()

            eq_(['TODO: do task1 (03f9)'], adapter._lines)

            # but it does not reminds twice
            now.return_value = datetime.datetime(2012, 9, 5, 2, 12)

            adapter._lines[:] = []
            plugin._remind_users_about_their_tasks()
            eq_([], adapter._lines)


def test_todo_done():
    with closing(Bot(adapters=[TestAdapter], plugins=[todo.Plugin])) as bot:

        adapter = bot.get_adapter('test')


        adapter.write('set timezone Asia/Shanghai')
        adapter.write('remind at 2012-09-05 10:00 to do task1')
        adapter.write('remind at 2012-09-05 10:30 to do task2')
        adapter.write('03 done')

        adapter._lines[:] = []
        adapter.write('my tasks')

        eq_(
            [
                '26) 2012-09-05 10:30 do task2',
            ],
           adapter._lines
        )


def test_todo_remind_at_uses_timezones():
    with closing(Bot(adapters=[TestAdapter], plugins=[todo.Plugin])) as bot:
        adapter = bot.get_adapter('test')
        plugin = bot.get_plugin('todo')
        identity_plugin = bot.get_plugin('identity')
        identity = identity_plugin.get_identity_by_user(adapter, User('some user'))

        adapter.write('set timezone Asia/Shanghai')
        adapter.write('remind at 2012-09-05 00:00 to do task1')
        tasks = plugin._get_tasks(identity)
        eq_(datetime.datetime(2012, 9, 4, 16, 0), tasks[0][0])


def test_yaml_config():
    cfg = Config()
    cfg.read_from_string("""
adapters: [xmpp, irc]
irc:
    channels: [thebot, test]
xmpp:
    jid: thebot@ya.ru
""")

    eq_(['xmpp', 'irc'], cfg.adapters)
    eq_('thebot@ya.ru', cfg.xmpp_jid)


def test_get_config_values_from_nested_structures():
    cfg = Config()
    cfg.read_from_string("""
xmpp:
    jid: thebot@ya.ru
    notify:
       adapter: irc
       to_username: art
       
""")
    eq_('thebot@ya.ru', cfg.xmpp['jid'])
    eq_('thebot@ya.ru', cfg.xmpp_jid)

    eq_('irc', cfg.xmpp_notify_adapter)
    eq_('art', cfg.xmpp_notify_to_username)


def _get_identity_id(adapter):
    match = re.match(
        r'.*bind to ([0-9a-f]{40}).*',
        adapter._lines[-1]
    )
    assert match is not None
    return match.group(1)


def test_identity_create():
    with closing(Bot(adapters=[TestAdapter], plugins=['identity'])) as bot:
        adapter = bot.get_adapter('test')

        adapter.write('build identity', user='user1')
        identity = _get_identity_id(adapter)

        adapter.write('bind to {}'.format(identity), user='user2')

        adapter._lines[:] = []
        adapter.write('show my accounts', user='user1')

        eq_(
            'Your identities are:\n'
            '1) user1 (test, online)\n'
            '2) user2 (test, online)',
            adapter._lines[0]
        )


def test_identity_get_by_xxx():
    with closing(Bot(adapters=[TestAdapter], plugins=['identity'])) as bot:
        adapter = bot.get_adapter('test')
        plugin = bot.get_plugin('identity')

        adapter.write('build identity', user='user1')
        identity = _get_identity_id(adapter)

        adapter.write('bind to {}'.format(identity), user='user2')

        eq_(identity, plugin.get_identity_by_id(identity).id)
        eq_(identity, plugin.get_identity_by_user(adapter, User('user1')).id)
        eq_(identity, plugin.get_identity_by_user(adapter, User('user2')).id)

        eq_(None, plugin.get_identity_by_id('unexistent'))

        # for the new user, a new identity will be created
        new_identity = plugin.get_identity_by_user(adapter, User('unknown user'))
        assert new_identity.id != identity


def test_identity_show_my_identities_when_there_is_no_identity():
    with closing(Bot(adapters=[TestAdapter], plugins=['identity'])) as bot:
        adapter = bot.get_adapter('test')

        adapter._lines[:] = []
        adapter.write('show my accounts', user='user1')

        eq_(
            'Your identities are:\n'
            '1) user1 (test, online)',
            adapter._lines[0]
        )

def test_bind_to_another_identity():
    """Testing bind to another identity, when there is already exists another identity."""
    with closing(Bot(adapters=[TestAdapter], plugins=['identity'])) as bot:
        adapter = bot.get_adapter('test')
        plugin = bot.get_plugin('identity')

        adapter._lines[:] = []
        adapter.write('show my accounts', user='user1')

        eq_(
            'Your identities are:\n'
            '1) user1 (test, online)',
            adapter._lines[-1]
        )

        adapter.write('show my accounts', user='user2')

        eq_(
            'Your identities are:\n'
            '1) user2 (test, online)',
            adapter._lines[-1]
        )

        # there should be two identities now
        eq_(2, len(plugin.identities))
        # and two persons
        eq_(2, len(plugin.persons))

        # now let's bind user2 to user1
        adapter.write('build identity', user='user1')
        identity = _get_identity_id(adapter)

        adapter.write('bind to {}'.format(identity), 'user2')

        # now there should be only one identity
        eq_(1, len(plugin.identities))
        # and still two persons
        eq_(2, len(plugin.persons))
        # pointing to the same identity
        eq_(
            [identity, identity],
            list(plugin.persons.values())
        )

def test_persons_equality():
    with closing(Bot(adapters=[TestAdapter], plugins=['identity'])) as bot:
        adapter = bot.get_adapter('test')
        eq_(
            Person(adapter, User('tester')),
            Person(adapter, User('tester')),
        )


def test_users_equality():
    eq_(
        User('tester'),
        User('tester'),
    )


def test_unbind_from_identity():
    with closing(Bot(adapters=[TestAdapter], plugins=['identity'])) as bot:
        adapter = bot.get_adapter('test')

        adapter.write('build identity', user='user1')
        identity = _get_identity_id(adapter)
        adapter.write('bind to {}'.format(identity), user='user2')
        adapter.write('unbind', user='user2')

        adapter._lines[:] = []
        adapter.write('show my accounts', user='user1')

        eq_(
            'Your identities are:\n'
            '1) user1 (test, online)',
            adapter._lines[0]
        )


def test_notify_first_online_person():
    with closing(Bot(adapters=[TestAdapter], plugins=['notify'])) as bot:
        adapter = bot.get_adapter('test')
        plugin = bot.get_plugin('notify')

        adapter.write('build identity', user='user1')
        identity_id = _get_identity_id(adapter)
        adapter.write('bind to {}'.format(identity_id), user='user2')
        # now we have identity with two users

        adapter._lines[:] = []

        # first, check if first user will receive message
        eq_(True, plugin.notify(identity_id, 'hello 1'))
        eq_('hello 1', adapter._lines[-1])

        adapter.offline('user1')

        eq_(True, plugin.notify(identity_id, 'hello 2'))
        eq_('hello 2', adapter._lines[-1])

        adapter.offline('user2')

        eq_(2, len(adapter._lines))
        eq_(False, plugin.notify(identity_id, 'hello 3'))
        # no messages was sent, because all persons are offline
        eq_(2, len(adapter._lines))


def test_notification_priorities():
    class TestAdapter2(TestAdapter):
        name = 'test2'

    with closing(Bot(adapters=[TestAdapter, TestAdapter2], plugins=['notify'])) as bot:
        adapter1 = bot.get_adapter('test')
        adapter2 = bot.get_adapter('test2')

        plugin = bot.get_plugin('notify')

        adapter1.write('build identity', user='user1')
        identity_id = _get_identity_id(adapter1)
        adapter2.write('bind to {}'.format(identity_id), user='user2')
        # now we have identity with two users

        adapter1._lines[:] = []
        adapter2._lines[:] = []

        # without priorities, first user will receive message
        eq_(True, plugin.notify(identity_id, 'hello 1'))
        eq_('hello 1', adapter1._lines[-1])

        adapter1.write('set notification-priorities test2,test', user='user1')

        # now second adapter has higher priority
        eq_(True, plugin.notify(identity_id, 'hello 2'))
        eq_('hello 2', adapter2._lines[-1])


def test_set_get_settings():
    with closing(Bot(adapters=[TestAdapter], plugins=['settings', 'identity'])) as bot:
        adapter = bot.get_adapter('test')
        plugin = bot.get_plugin('settings')

        adapter.write('build identity', user='user1')
        identity_id = _get_identity_id(adapter)
        adapter.write('bind to {}'.format(identity_id), user='user2')
        # now we have identity with two users

        plugin.set(identity_id, 'some-key', 'blah')
        eq_('blah', plugin.get(identity_id, 'some-key'))

        adapter.write('set some-key some-value', user='user1')
        eq_('some-value', plugin.get(identity_id, 'some-key'))

        plugin.set(identity_id, 'some-key', 'another-value')
        adapter.write('get some-key', user='user2')
        eq_('some-key = another-value', adapter._lines[-1])

        adapter.write('my settings', user='user1')
        eq_('You settings are:\nsome-key = another-value', adapter._lines[-1])


def test_load_dependencies():
    """Plugin notify should depend on settings, and settings plugin in it's
    turn, depends on identity.
    """

    with closing(Bot(adapters=[], plugins=['notify'])) as bot:
        settings = bot.get_plugin('settings')
        assert settings is not None

        identity = bot.get_plugin('identity')
        assert identity is not None

def test_stub_methods():
    stub = Stub('blah')
    eq_('blah', stub.name)
    eq_('blah.is_online', '{}'.format((stub.is_online)))
    eq_(None, stub.is_online())




# coding: utf-8
from __future__ import absolute_import, unicode_literals

import bisect
import times
import hashlib
import six

from dateutil.parser import parse
from thebot import ThreadedPlugin, on_command


def _gen_hashes(tasks):
    """Generates sha1 hashes for tasks and minumum subhash's
    length for which all generated hashes are unique.
    """
    if not tasks:
        return [], 0

    shas = [
        hashlib.sha1(about.encode('utf-8')).hexdigest()
        for dt, about, r in tasks
    ]
    for min_len in range(2, len(shas[0]) + 1):
        trimmed = set(h[:min_len] for h in shas)
        if len(trimmed) == len(shas):
            break
    return shas, min_len


class Plugin(ThreadedPlugin):
    """Allows to manage a simple todo list.

    By default, all dates are in the UTC.
    To use your local time everywhere, set timezone like that:

    set timezone Europe/Moscow
    """
    name = 'todo'
    deps = ['notify', 'identity', 'settings']

    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)

        # interval between bot's attemts to find expired tasks
        self.interval = 60

        if not self.bot.config.unittest:
            self.start_worker(interval=self.interval)

    def _get_tasks(self, identity):
        return self.storage.get('tasks:{}'.format(identity.id), [])

    def _set_tasks(self, identity, tasks):
        self.storage['tasks:{}'.format(identity.id)] = tasks

    @on_command('remind( me)? at (?P<datetime>.+) to (?P<about>.+)')
    def remind(self, request, datetime, about):
        """Remind about a TODO at given time."""

        try:
            identity = self.bot.get_plugin('identity').get_identity_by_request(request)
            dt = parse(datetime)
            tz = self._get_user_timezone(identity)
            dt = times.to_universal(dt, tz)

            tasks = self._get_tasks(identity)
            bisect.insort(tasks, (dt, about, identity.id))
            self._set_tasks(identity, tasks)
        except Exception as e:
            request.respond('Unable to parse a date: ' + six.text_type(e))
            raise

        request.respond('ok')

    @on_command('my tasks')
    def my_tasks(self, request):
        """Show my tasks"""
        identity = self.bot.get_plugin('identity').get_identity_by_request(request)
        tasks = self._get_tasks(identity)

        if tasks:
            tz = self._get_user_timezone(identity)
            hashes, min_len = _gen_hashes(tasks)

            lines = []
            for h, (dt, about, identity_id) in zip(hashes, tasks):
                dt = times.to_local(dt, tz)
                lines.append('{0}) {1:%Y-%m-%d %H:%M} {2}'.format(h[:min_len], dt, about))

            request.respond('\n'.join(lines))
        else:
            request.respond('You have no tasks')

    @on_command('(?P<task_id>[0-9a-z]{2,40}) done')
    def done(self, request, task_id):
        """Mark given task as done."""
        # TODO add unittests for two cases:
        # * when task_id not found
        # * when there are more than one task which hash started from task_id
        identity = self.bot.get_plugin('identity').get_identity_by_request(request)
        tasks = self._get_tasks(identity)

        if tasks:
            hashes, min_len = _gen_hashes(tasks)
            filtered_tasks = []
            for h, (dt, about, identity_id) in zip(hashes, tasks):
                if not h.startswith(task_id):
                    filtered_tasks.append((dt, about, identity_id))

            self._set_tasks(identity, filtered_tasks)
        request.respond('done')

    def _remind_users_about_their_tasks(self):
        now = times.now()
        tasks_storage = self.storage.with_prefix('tasks:')
        notify = self.bot.get_plugin('notify').notify

        for identity_id, todos in tasks_storage.items():
            idx = bisect.bisect_left(todos, (now, None, None))
            to_remind = todos[:idx]

            for td, about, id_id in to_remind:
                delta = (now - td)
                if delta.seconds <= self.interval:
                    # Remind only if reminder's datetime is between
                    # this and previous checks
                    notify(identity_id, 'TODO: {0} ({1})'.format(
                        about,
                        hashlib.sha1(about.encode('utf-8')).hexdigest()[:4]
                    ))

    def do_job(self):
        self._remind_users_about_their_tasks()

    def _get_user_timezone(self, identity):
        settings = self.bot.get_plugin('settings')
        return settings.get(identity.id, 'timezone', 'UTC')

    @on_command('now')
    def now(self, request):
        """Outputs server time and user time."""
        identity = self.bot.get_plugin('identity').get_identity_by_request(request)

        now = times.now()
        tz = self._get_user_timezone(identity)
        local = times.to_local(now, tz)

        request.respond('Server time: {}\nLocal time:{}'.format(now, local))




# coding: utf-8
from __future__ import absolute_import, unicode_literals

import hashlib
import time
import random

from thebot import Plugin, on_command
from thebot.utils import printable


@printable
class Person(object):
    def __init__(self, adapter, user):
        self.adapter = adapter
        self.user = user

    def __unicode__(self):
        return '{} ({})'.format(self.user, self.adapter)

    def __eq__(self, another):
        return self.adapter == another.adapter and self.user == another.user


class Identity(object):
    def __init__(self, identity_id):
        self.id = identity_id
        self.persons = []


class Plugin(Plugin):
    """Allows to join several accounts from different chats.

    This plugin is used by other plugins, to retrive a user's UID.
    You can join several accounts from different adapters, and
    all of them will be binded to the same UID.

    That way, user's information can be accessible from different
    instant messengers.
    """
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.identities = {}
        self.persons = {} # a map from person to identity.id

        self.identity_storage = self.storage.with_prefix('i:')

        for identity in self.identity_storage.values():
            self._add_identity(identity, save_to_storage=False)

    @on_command('build identity')
    def build(self, request):
        """Extend you identity with another accounts."""
        identity = self.get_identity_by_user(request.adapter, request.user)

        request.respond(
            r'Ok, please, send me this command via other adapters: "bind to {}"'.format(identity.id)
        )

    def _add_identity(self, identity, save_to_storage=True):
        if save_to_storage:
            self.identity_storage[identity.id] = identity

        self.identities[identity.id] = identity
        for person in identity.persons:
            self.persons[(person.adapter.name, person.user.id)] = identity.id

    def _create_identity(self, adapter, user):
        identity = Identity(
            hashlib.sha1(
                (str(time.time()) + str(random.random())).encode('utf-8')
            ).hexdigest()
        )
        identity.persons.append(Person(adapter, user))
        self._add_identity(identity)

        self.logger.debug('Created identity {}'.format(identity.id))
        return identity

    @on_command('bind to (?P<identity_id>[0-9a-f]{40})')
    def bind(self, request, identity_id):
        """Bind current account to a given identity id. To use this command, execute 'build identity' from another account first."""
        to_identity = self.identities.get(identity_id, None)

        if to_identity is None:
            request.respond('Identity with id {} not found'.format(identity_id))
        else:
            from_identity = self.get_identity_by_request(request)

            if from_identity != to_identity:
                person = Person(request.adapter, request.user)
                self.logger.debug('Binding {} to identity {}'.format(person, to_identity.id))

                to_identity.persons.append(person)
                self._add_identity(to_identity)

                from_identity.persons.remove(person)
                if len(from_identity.persons) == 0:
                    del self.identity_storage[from_identity.id]
                    del self.identities[from_identity.id]
            request.respond('ok')

    @on_command('unbind')
    def unbind(self, request):
        """Unbind current account from the identity."""
        from_identity = self.get_identity_by_request(request)
        person = Person(request.adapter, request.user)

        self.logger.debug('Unbinding {} to identity {}'.format(person, from_identity.id))
        from_identity.persons.remove(person)

        if len(from_identity.persons) == 0:
            del self.identity_storage[from_identity.id]
            del self.identities[from_identity.id]

    @on_command('show my accounts')
    def show_my_ids(self, request):
        """Show all accounts, binded to a current identity."""
        identity = self.get_identity_by_request(request)
        if identity is not None:
            # TODO rename identitites to accounts
            request.respond(
                'Your identities are:\n' + '\n'.join(
                    '{}) {} ({}, {})'.format(
                        idx,
                        contact.user,
                        contact.adapter,
                        'online' if contact.adapter.is_online(contact.user) else 'offline'
                    )
                    for idx, contact in enumerate(identity.persons, 1)
                )
            )

    def get_identity_by_id(self, identity_id):
        return self.identities.get(identity_id)

    def get_identity_by_user(self, adapter, user):
        """Returns identity for user.

        If it does not exist, then identity will be created.
        """
        identity_id = self.persons.get((adapter.name, user.id), None)
        if identity_id is None:
            return self._create_identity(adapter, user)
        else:
            return self.identities.get(identity_id, None)

    def get_identity_by_request(self, request):
        return self.get_identity_by_user(request.adapter, request.user)




from __future__ import absolute_import, unicode_literals, print_function

import email
import email.message
import imaplib
import smtplib
import logging
import time
import thebot
import threading

#from email.mime.text import MIMEText


class Request(thebot.Request):
    def __init__(self, adapter, message, from_email, message_id, subject):
        super(Request, self).__init__(adapter, message, user=thebot.User(from_email))
        self.message_id = message_id
        self.subject = subject

    def respond(self, message):
        subject = self.subject

        # if there wasn't any message, then probably this request
        # was created just to notify somebody on some event
        if self.message and not subject.lower().startswith('re: '):
            subject = 'Re: ' + subject

        citate = ''
        if self.message:
            citate = '> {}\n'.format(self.message)
            
        self.adapter.send(
            citate + message,
            self.user,
            subject=subject,
            in_reply_to=self.message_id,
        )

    def shout(self, message):
        self.respond(message)


def _get_text_from_email(message):

    def _get_text(message):
        if message.is_multipart():
            for part in message.get_payload():
                text = _get_text(part)
                if text is not None:
                    return text
        else:
            mimetype = message.get_content_type()
            if mimetype == 'text/plain':
                payload = message.get_payload()
                charset = message.get_content_charset()
                if charset is not None:
                    payload = payload.decode(charset, 'replace')
                return payload

    return _get_text(message)


class Imap(object):
    def __init__(self, host, port, username, password):
        self.logger = logging.getLogger('thebot.batteries.mail.imap')
        if port == 143:
            self._imap = imaplib.IMAP4(host, port)
        else:
            self._imap = imaplib.IMAP4_SSL(host, port)
        self.username = username
        self.password = password

    def __enter__(self):
        status, messages = self._imap.login(self.username, self.password)
        if status != 'OK':
            raise RuntimeError(''.join(messages))

        status, messages = self._imap.select()
        if status != 'OK':
            raise RuntimeError(' '.join(messages))

        return self

    def __exit__(self, *args):
        self._imap.logout()

    def fetch_messages(self):
        status, data = self._imap.uid('search', None, 'ALL')
        uids = data[0].split() if data[0] is not None else []

        for uid in uids:
            self.logger.debug('fetching message with uid {}'.format(uid))
            status, data = self._imap.uid('fetch', uid, '(RFC822)')
            message = email.message_from_string(data[0][1])
            message.uid = uid
            yield message

    def delete(self, message):
        self.logger.debug('deleting message with uid {}'.format(message.uid))
        self._imap.uid('store', message.uid, '+FLAGS', '\\Deleted')
        self._imap.expunge()


class Adapter(thebot.Adapter):
    @staticmethod
    def get_options(parser):
        group = parser.add_argument_group('SMTP options')
        group.add_argument(
            '--smtp-host', default='localhost',
            help='SMTP server to connect. Default: localhost.'
        )
        group.add_argument(
            '--smtp-port', default=25,
            help='SMTP port to connect. Default: 25.'
        )
        group.add_argument(
            '--smtp-username',
            help='Username to connect to SMTP server.'
        )
        group.add_argument(
            '--smtp-password',
            help='Password to connect to SMTP server.'
        )
        group.add_argument(
            '--smtp-from',
            help='An email to use in From: header.'
        )

        group = parser.add_argument_group('IMAP options')
        group.add_argument(
            '--imap-host', default='localhost',
            help='IMAP server to connect. Default: localhost.'
        )
        group.add_argument(
            '--imap-port', default=143,
            help='IMAP port to connect. Default: 143.'
        )
        group.add_argument(
            '--imap-username',
            help='Username to connect to IMAP server.'
        )
        group.add_argument(
            '--imap-password',
            help='Password to connect to IMAP server.'
        )


    def start(self):
        thread = threading.Thread(target=self._fetch_messages)
        thread.daemon = True
        thread.start()

    def get_imap(self):
        cfg = self.bot.config
        return Imap(cfg.imap_host, int(cfg.imap_port), cfg.imap_username, cfg.imap_password)

    def create_request(self,
                       message=None,
                       email=None,
                       message_id=None,
                       subject=None):
        return Request(self, message, email, message_id, subject)

    def _fetch_messages(self):
        logger = logging.getLogger('thebot.batteries.mail')

        while True:
            with self.get_imap() as imap:
                for message in imap.fetch_messages():
                    try:
                        from_ = email.Header.decode_header(message['from'])
                        full_from = (value.decode(charset or 'ascii') for value, charset in from_)
                        full_from = ' '.join(full_from)
                        from_email = [value for value, charset in from_ if charset is None and '@' in value][0]

                        text = _get_text_from_email(message)
                        if text is None:
                            logger.warning('Message from {} contains no plain/text part.'.format(full_from))
                        else:
                            # selecting first non empty line
                            lines = text.split('\n')
                            lines = (line.strip() for line in lines)
                            lines = list(filter(None, lines))
                            first_line = lines[0]
                            request = self.create_request(
                                message=first_line,
                                email=from_email,
                                message_id=message['Message-Id'],
                                subject=message['Subject'],
                            )
                            self.callback(request, direct=True)

                            imap.delete(message)
                    except Exception:
                        logger.exception('processing message')

                time.sleep(5)

    def send(self, message, user, subject='Message from TheBot', in_reply_to=None):
        to_email = user.id

        logger = logging.getLogger('thebot.batteries.mail')
        logger.debug('Sending email to "{}" in reply to "{}"'.format(to_email, in_reply_to))

        cfg = self.bot.config

        port = int(cfg.smtp_port)
        if port == 25:
            server = smtplib.SMTP(cfg.smtp_host, port)
        else:
            server = smtplib.SMTP_SSL(cfg.smtp_host, port)

        server.login(cfg.smtp_username, cfg.smtp_password)
        #server.set_debuglevel(1)

        from_email = getattr(cfg, 'smtp_from', None)
        if from_email is None and '@' in cfg.smtp_username:
            from_email = cfg.smtp_username
        else:
            raise RuntimeError('Please, specify "--smtp-from" option.')

        response = email.message.Message()
        response['From'] = 'dev.thebot@ya.ru'
        response['To'] = to_email
        response['Subject'] = subject
        if in_reply_to:
            response['In-Reply-To'] = in_reply_to
        response.set_payload(message.encode('utf8'), 'utf-8')

        server.sendmail(from_email, to_email, response.as_string())
        server.quit()

    def close(self):
        self.imap.close()
        self.imap.logout()





from __future__ import absolute_import, unicode_literals

import sys

try:
    from collections import MutableMapping
except ImportError:
    from UserDict import DictMixin as MutableMapping


if sys.version_info.major == 3:
    force_unicode = lambda x: x
    force_str = lambda x: x
else:
    def force_unicode(value):
        if isinstance(value, str):
            value = value.decode('utf-8')
        return value

    def force_str(value):
        if isinstance(value, unicode):
            value = value.encode('utf-8')
        return value


def printable(cls):
    """This decorator creates __str__ object for python2.x
    or renames __uncode__ into __str__ for python3.x
    """
    if sys.version_info.major == 3:
        setattr(cls, '__str__', getattr(cls, '__unicode__'))
        delattr(cls, '__unicode__')
    else:
        setattr(cls, '__str__', lambda self: self.__unicode__().encode('utf8'))

    return cls




from __future__ import absolute_import, unicode_literals

import re
import irc
import logging
import time
import thebot
import threading

from collections import defaultdict
from functools import wraps


class IRCConnection(irc.IRCConnection):
    def __init__(self, *args, **kwargs):
        super(IRCConnection, self).__init__(*args, **kwargs)
        # a map nick -> list of events
        # to signal about received online status
        self._ison_requests = defaultdict(list)
        # a map nick -> True/False
        # if True, then this nick is online
        self._online_statuses = {}

    def get_logger(self, logger_name, filename):
        """We override this method because don't want to have a separate log for irc messages.
        """
        return logging.getLogger(logger_name)

    def dispatch_patterns(self):
        callbacks = list(super(IRCConnection, self).dispatch_patterns())
        callbacks.append(
            (re.compile(':(?:.*?)\s+303(?:.*?):(?P<nicks>.*)'), self.on_ison_response)
        )
        def threaded_callback(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                thread = threading.Thread(target=func, args=args, kwargs=kwargs)
                thread.daemon = True
                thread.start()
            return wrapper

        callbacks = tuple(
            (pattern, threaded_callback(callback))
            for pattern, callback
                in callbacks
        )
        return callbacks

    def is_online(self, nick):
        """This method sends ISON request and waits for response."""
        event = threading.Event()
        self._ison_requests[nick].append(event)
        self._online_statuses[nick] = False

        self.send('ISON {}'.format(nick))
        event.wait(timeout=3)
        return self._online_statuses[nick]

    def on_ison_response(self, nicks):
        logger = logging.getLogger('thebot.adapter.irc')
        logger.debug('These nicks are online: ' + nicks)

        for nick in nicks.split(' '):
            self._online_statuses[nick] = True
            for event in self._ison_requests[nick]:
                event.set()


class Adapter(thebot.Adapter):
    @staticmethod
    def get_options(parser):
        group = parser.add_argument_group('IRC options')
        group.add_argument(
            '--irc-host', default='irc.freenode.net',
            help='Server to connect. Default: irc.freenode.net.'
        )
        group.add_argument(
            '--irc-port', default=6667, type=int,
            help='Port to connect. Default: 6667.'
        )
        group.add_argument(
            '--irc-channels', default='thebot',
            help='Comma-separated list of channels. Default: thebot.',
        )
        group.add_argument(
            '--irc-nick', default='thebot',
            help='IRC nick. Default: thebot.',
        )


    def start(self):
        thread = threading.Thread(target=self.run_bot)
        thread.daemon = True
        thread.start()

    def run_bot(self):
        """
        Convenience function to start a bot on the given network, optionally joining
        some channels
        """
        host = self.bot.config.irc_host
        port = self.bot.config.irc_port
        nick = self.bot.config.irc_nick
        channels = self.bot.config.irc_channels.split(',')

        def on_message(nick, message, channel):
            """A callback to be called by irckit when new message will arrive.

            It verifies if a bot's nick is mentioned, and pass data to TheBot.
            """
            logging.getLogger('adapter.irc').debug(
                'Received message "{}" from {} at channel {}.'.format(
                    message, nick, channel
                )
            )
            message = thebot.utils.force_unicode(message)
            nick_re = re.compile('^%s[:,\s]\s*' % conn.nick)

            direct = nick_re.match(message) is not None
            message = nick_re.sub('', message)

            request = thebot.Request(
                self,
                message,
                thebot.User(nick),
                thebot.Room(channel) if channel else None,
                refer_by_name=direct
            )
            return self.callback(request, direct=direct)

        conn = IRCConnection(host, port, nick)
        conn.register_callbacks((
            (re.compile('.*'), on_message),
        ))

        self.irc_connection = conn

        while 1:
            conn.connect()
            channels = channels or []
            for channel in channels:
                conn.join(channel)
            conn.enter_event_loop()

    def send(self, message, user=None, room=None, refer_by_name=False):
        logger = logging.getLogger('thebot.adapter.irc')

        sleep_time = 0.05
        max_sleep_time = 1

        for line in message.split('\n'):
            if refer_by_name:
                line = user.id + ', ' + line

            logger.debug('Sending "{}" to {} at {}'.format(
                line,
                'everybody' if user is None else user,
                'private channel' if room is None else room,
            ))
            self.irc_connection.respond(
                thebot.utils.force_str(line),
                nick=user.id if user else None,
                channel=room.id if room else None,
            )

            # this is a protection from the flood
            # if we'll send to often, then server may decide to
            # logout us
            time.sleep(min(sleep_time, max_sleep_time))
            sleep_time *= 2

    def is_online(self, user):
        return self.irc_connection.is_online(user.id)




from __future__ import absolute_import, unicode_literals

import irc
import logging
import threading

from wsgiref.simple_server import make_server, WSGIRequestHandler
from .. import Request, Adapter, User, __version__
from ..utils import force_str
from cgi import parse_qs

class HttpRequest(Request):
    def __init__(self, adapter, environ, start_response):
        super(HttpRequest, self).__init__(adapter, environ['PATH_INFO'], user=User('http service'))
        self.environ = environ
        self.start_response = start_response
        self.response_sent = False
        self.method = environ['REQUEST_METHOD']

        self.GET = parse_qs(environ['QUERY_STRING'])

        if self.method == 'POST':
            content_length = int(environ.get('CONTENT_LENGTH', 0))
            request_body = environ['wsgi.input'].read(content_length)
            self.POST = parse_qs(request_body)
        else:
            self.POST = None

    def respond(self, message):
        if self.response_sent:
            raise RuntimeError('Response to this HTTP request already sent.')

        status = b'200 OK'
        headers = [
            (b'Server', b'TheBot/' + force_str(__version__)),
            (b'Content-type', b'text/plain; charset=utf-8'),
        ]

        write = self.start_response(status, headers)
        write(force_str(message))
        self.response_sent = True


class Adapter(Adapter):
    @staticmethod
    def get_options(parser):
        group = parser.add_argument_group('HTTP options')
        group.add_argument(
            '--http-host', default='127.0.0.1',
            help='IP to bind to. Default: 127.0.0.1.'
        )
        group.add_argument(
            '--http-port', default=8888,
            help='TCP port to bind to. Default: 8888.'
        )


    def start(self):
        class QuietHandler(WSGIRequestHandler):
            def __init__(self, *args, **kwargs):
                self.logger = logging.getLogger('thebot.batteries.http')
                WSGIRequestHandler.__init__(self, *args, **kwargs)

            def log_message(self, format, *args):
                self.logger.info(
                    '%s - - [%s] %s',
                    self.address_string(),
                    self.log_date_time_string(),
                    format % args
                )

        server = make_server(
            self.bot.config.http_host,
            int(self.bot.config.http_port),
            self._wsgi_handler,
            handler_class=QuietHandler,
        )

        thread = threading.Thread(target=server.serve_forever)
        thread.daemon = True
        thread.start()

    def _wsgi_handler(self, environ, start_response):
        request = HttpRequest(self, environ, start_response)
        self.callback(request)
        if not request.response_sent:
            request.respond('')
        return []




# coding: utf-8
from __future__ import absolute_import, unicode_literals

import sleekxmpp
import thebot
import threading


class User(thebot.User):
    def __init__(self, jid):
        self.id, self.resource = jid.split('/')


class Adapter(thebot.Adapter):
    ignore_jids = [
        'lastmail.ya.ru/Яндекс.Информер',
    ]
    @staticmethod
    def get_options(parser):
        group = parser.add_argument_group('XMPP options')
        group.add_argument(
            '--xmpp-jid', default='thebot@ya.ru',
            help='Jabber JID. Default: thebot@ya.ru.'
        )
        group.add_argument(
            '--xmpp-server', default='',
            help='Jabber server. Optional.'
        )
        group.add_argument(
            '--xmpp-password', default='',
            help='Password to connect to the server. Default: "".'
        )

    def start(self):
        thread = threading.Thread(target=self.run_bot)
        thread.daemon = True
        thread.start()

    def run_bot(self):
        """
        Convenience function to start a bot on the given network, optionally joining
        some channels
        """

        jid = self.bot.config.xmpp_jid
        password = self.bot.config.xmpp_password

        def on_message(msg):
            """A callback to be called by xmpppy's when new message will arrive.

            In it's turn, it will call TheBot's callback, to pass request into it.
            """

            if msg['type'] in ('chat', 'normal'):
                if msg['from'] != msg['to'] and msg['from'] not in self.ignore_jids:
                    request = thebot.Request(
                        self,
                        msg['body'],
                        user=User(unicode(msg['from']))
                    )
                    self.callback(request)


        def on_start(event):
            self.xmpp_bot.get_roster()
            self.xmpp_bot.send_presence()


        self.xmpp_bot = sleekxmpp.ClientXMPP(jid, password)
        self.xmpp_bot._use_daemons = True
        self.xmpp_bot.add_event_handler('session_start', on_start)
        self.xmpp_bot.add_event_handler('message', on_message)

        self.xmpp_bot.connect()
        self.xmpp_bot.process(block=True)

    def send(self, message, user=None, room=None, refer_by_name=False):
        msg = sleekxmpp.stanza.message.Message()
        msg['to'] = user.id + '/' + user.resource
        msg['type'] = 'chat'
        msg['body'] = message

        self.xmpp_bot.send(msg)

    def is_online(self, user):
        return len(self.xmpp_bot.client_roster.presence(user.id)) > 0




import os.path
import sys

from setuptools import setup, find_packages


if sys.version > '3':
    req_filename = 'requirements/production3.txt'
else:
    req_filename = 'requirements/production.txt'

if os.path.exists(req_filename):
    requirements = open(req_filename).readlines()
else:
    # this branch should work only when running under the tox
    requirements = []


setup(
    name='thebot',
    version='0.4.1',
    description=(
    ),
    keywords='chat irc xmpp basecamp jira fun',
    license = 'New BSD License',
    author="Alexander Artemenko",
    author_email='svetlyak.40wt@gmail.com',
    url='http://github.com/svetlyak40wt/thebot/',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    packages=find_packages(),
    scripts=['scripts/thebot'],
    install_requires=requirements,
)



# coding: utf-8
from __future__ import absolute_import, unicode_literals

from thebot import Plugin, on_command


class Plugin(Plugin):
    """Stores user settings."""
    deps = ['identity']

    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.user_settings = self.storage.with_prefix('user:')

    @on_command('set (?P<key>[a-z][a-z-]*) (?P<value>.*)')
    def on_set(self, request, key, value):
        """Set a key to a given value."""
        identity_plugin = self.bot.get_plugin('identity')
        identity = identity_plugin.get_identity_by_user(request.adapter, request.user)
        self.set(identity.id, key, value)

    @on_command('get (?P<key>[a-z][a-z-]*)')
    def on_get(self, request, key):
        """Get setting with a given key."""
        identity_plugin = self.bot.get_plugin('identity')
        identity = identity_plugin.get_identity_by_user(request.adapter, request.user)
        try:
            request.respond('{} = {}'.format(
                key,
                self.get(identity.id, key)
            ))
        except KeyError:
            request.respond('Settings {} not found'.format(key))

    @on_command('my settings')
    def my_settings(self, request):
        """Show all settings."""
        identity_plugin = self.bot.get_plugin('identity')
        identity = identity_plugin.get_identity_by_user(request.adapter, request.user)
        user_settings = self.user_settings.with_prefix('{}:'.format(identity.id))

        lines = []
        for key, value in user_settings.items():
            lines.append('{} = {}'.format(key, value))

        if lines:
            lines.insert(0, 'You settings are:')
        else:
            lines.append('You have no any settings yet.')

        request.respond('\n'.join(lines))

    def get(self, identity_id, key, default=None):
        return self.user_settings.get('{}:{}'.format(identity_id, key), default)

    def set(self, identity_id, key, value):
        self.user_settings['{}:{}'.format(identity_id, key)] = value




from __future__ import absolute_import, unicode_literals

import sys
import threading
import thebot


class Adapter(thebot.Adapter):
    def start(self):
        def loop():
            while True:
                sys.stdout.write('> ')
                sys.stdout.flush()
                line = sys.stdin.readline()
                if len(line) == 0:
                    # seems, Ctrl-D was pressed
                    self.callback(thebot.EXIT)
                    return

                line = line.strip()

                request = thebot.Request(
                    self,
                    line,
                    thebot.User('console'),
                )
                self.callback(request)

        thread = threading.Thread(target=loop)
        thread.daemon = True
        thread.start()

    def send(self, message, user=None, room=None, refer_by_name=False):
        sys.stdout.write('{0}\n'.format(message))
        sys.stdout.flush()

    def is_online(self, user):
        return True
